from __future__ import division, print_function

import unittest

import numpy
from foldedleastsquares import cleaned_array

class TestCleanedArray(unittest.TestCase):
    def test(self):
        print("Starting test: cleaned_array...", end="")

        dirty_array = numpy.ones(10, dtype=object)
        time_array = numpy.linspace(1, 10, 10)
        dy_array = numpy.ones(10, dtype=object)
        dirty_array[1] = None
        dirty_array[2] = numpy.inf
        dirty_array[3] = -numpy.inf
        dirty_array[4] = numpy.nan
        dirty_array[5] = -99
        time_array[8] = numpy.nan
        dy_array[9] = numpy.inf

        t, y, dy = cleaned_array(time_array, dirty_array, dy_array)
        numpy.testing.assert_equal(t, [1, 7, 8])
        numpy.testing.assert_equal(y, [1, 1, 1])
        numpy.testing.assert_equal(dy, [1, 1, 1])
        print("passed")



from __future__ import division, print_function

import unittest

import numpy
import scipy.signal
from foldedleastsquares import transitleastsquares
from foldedleastsquares.template_generator.default_transit_template_generator import DefaultTransitTemplateGenerator
import pkg_resources


class TestTransitTemplateGenerator(DefaultTransitTemplateGenerator):
    def __init__(self):
        super().__init__()

class TestShapes(unittest.TestCase):
    def loadfile(self, filename):
        filename = TestShapes.get_path(filename)
        data = numpy.genfromtxt(filename, delimiter=",", dtype="f8, f8", names=["t", "y"])
        return data["t"], data["y"]
    @staticmethod
    def get_path(path):
        """
        Gets right path for tests environment
        :param path:
        :return: the real path of the test resource
        """
        return pkg_resources.resource_filename(__name__, path)

    def test(self):
        print("Starting test: transit shapes...", end="")

        # Testing transit shapes
        t, y = self.loadfile("EPIC206154641.csv")
        trend = scipy.signal.medfilt(y, 25)
        y_filt = y / trend

        # box
        model_box = transitleastsquares(t, y_filt)
        results_box = model_box.power(transit_template="box", period_grid=numpy.arange(1, 5, 0.01))
        numpy.testing.assert_equal(len(model_box.period_grid), 400)

        results_box = model_box.power(transit_template="box")
        numpy.testing.assert_almost_equal(
            results_box.duration, 0.06207881501022775, decimal=5)
        numpy.testing.assert_almost_equal(results_box.rp_rs, 0.08836981203437415, decimal=5)
        print("Test passed: Box-shaped")

        # grazing
        model_grazing = transitleastsquares(t, y_filt)
        results_grazing = model_grazing.power(transit_template="grazing")

        numpy.testing.assert_almost_equal(
            results_grazing.duration, 0.09079026695245815, decimal=5
        )
        numpy.testing.assert_almost_equal(
            min(results_grazing.chi2red), 0.06759475703796078, decimal=5)
        print("Test passed: Grazing-shaped")

        # comet
        model_comet = transitleastsquares(t, y_filt)
        results_comet = model_comet.power(transit_template="tailed")
        numpy.testing.assert_almost_equal(
            results_comet.duration, 0.23745146741412124, decimal=5
        )
        numpy.testing.assert_almost_equal(
            min(results_comet.chi2red), 0.0980794344892094, decimal=5)
        print("Test passed: Comet-shaped")

        model_custom = transitleastsquares(t, y_filt)
        try:
            results_custom = model_custom.power(transit_template="custom",
                                                transit_template_generator="wrongTransitTemplateGenerator")
            assert False
        except ValueError as e:
            if e.args[0] == "The custom transit_template_generator does not implement TransitTemplateGenerator.":
                print("Test passed: Wrong custom transit template generator.")
            else:
                assert False

        # custom
        model_custom = transitleastsquares(t, y_filt)
        results_custom = model_custom.power(transit_template="custom",
                                            transit_template_generator=TestTransitTemplateGenerator())
        numpy.testing.assert_almost_equal(
            results_custom.duration, 0.06828669651125058, decimal=5
        )
        numpy.testing.assert_almost_equal(
            min(results_custom.chi2red), 0.09977336183179186, decimal=5)
        print("Test passed: Custom-shaped")

        print("All tests passed")



from __future__ import division, print_function
import numpy
from foldedleastsquares import resample

import unittest

class TestResample(unittest.TestCase):
    def test(self):
        print("Starting test resample...", end="")

        # Resampling
        testtime = numpy.linspace(0, 1, 1000)
        testflux = numpy.linspace(0.99, 1.01, 1000)
        a, b = resample(time=testtime, flux=testflux, factor=100)
        expected_result_a = (
            0.0,
            0.11111111,
            0.22222222,
            0.33333333,
            0.44444444,
            0.55555556,
            0.66666667,
            0.77777778,
            0.88888889,
            1.0,
        )
        numpy.testing.assert_almost_equal(a, expected_result_a)
        expected_result_b = (
            0.99,
            0.99222222,
            0.99444444,
            0.99666667,
            0.99888889,
            1.00111111,
            1.00333333,
            1.00555556,
            1.00777778,
            1.01,
        )
        numpy.testing.assert_almost_equal(b, expected_result_b)
        numpy.testing.assert_equal(len(a), 10)
        numpy.testing.assert_almost_equal(min(a), 0)
        numpy.testing.assert_almost_equal(max(a), 1)
        numpy.testing.assert_almost_equal(numpy.mean(a), 0.5)
        numpy.testing.assert_equal(len(b), 10)
        numpy.testing.assert_almost_equal(min(b), 0.99)
        numpy.testing.assert_almost_equal(max(b), 1.01)
        numpy.testing.assert_almost_equal(numpy.mean(b), 1)
        print("passed")



from __future__ import division, print_function

import unittest

import numpy
from foldedleastsquares import catalog_info
import sys


class TestCatalog(unittest.TestCase):
    def test(self):
        if sys.version_info >= (3, 0):
            print("Starting test: catalog_data...")
            (a, b), mass, mass_min, mass_max, radius, radius_min, radius_max = catalog_info(
                EPIC_ID=204341806
            )
            numpy.testing.assert_almost_equal((a, b), (1.1605, -0.1704))
            numpy.testing.assert_almost_equal(mass, numpy.nan)
            numpy.testing.assert_almost_equal(mass_min, numpy.nan)
            numpy.testing.assert_almost_equal(mass_max, numpy.nan)
            numpy.testing.assert_almost_equal(radius, numpy.nan)
            numpy.testing.assert_almost_equal(radius_min, numpy.nan)
            numpy.testing.assert_almost_equal(radius_max, numpy.nan)

            (a, b), mass, mass_min, mass_max, radius, radius_min, radius_max = catalog_info(
                EPIC_ID=204099713
            )
            numpy.testing.assert_almost_equal((a, b), (0.4804, 0.1867))
            numpy.testing.assert_almost_equal(mass, 1.046)
            numpy.testing.assert_almost_equal(mass_min, 0.898)
            numpy.testing.assert_almost_equal(mass_max, 0.642)
            numpy.testing.assert_almost_equal(radius, 1.261)
            numpy.testing.assert_almost_equal(radius_min, 1.044)
            numpy.testing.assert_almost_equal(radius_max, 0.925)
            print("Test passed: EPIC catalog pull from Vizier using astroquery")

            (a, b), mass, mass_min, mass_max, radius, radius_min, radius_max = catalog_info(
                KIC_ID=757076
            )
            numpy.testing.assert_almost_equal((a, b), (0.5819, 0.137))
            numpy.testing.assert_almost_equal(mass, 1.357)
            numpy.testing.assert_almost_equal(mass_max, 0.204)
            numpy.testing.assert_almost_equal(mass_min, 0.475)
            numpy.testing.assert_almost_equal(radius, 3.128)
            numpy.testing.assert_almost_equal(radius_max, 0.987)
            numpy.testing.assert_almost_equal(radius_min, 2.304)

            (a, b), mass, mass_min, mass_max, radius, radius_min, radius_max = catalog_info(
                KIC_ID=12557548
            )
            numpy.testing.assert_almost_equal((a, b), (0.701, 0.0462))
            numpy.testing.assert_almost_equal(mass, 0.6666, decimal=3)
            numpy.testing.assert_almost_equal(mass_max, 0.067, decimal=3)
            numpy.testing.assert_almost_equal(mass_min, 0.055, decimal=3)
            numpy.testing.assert_almost_equal(radius, 0.660, decimal=3)
            numpy.testing.assert_almost_equal(radius_max, 0.054, decimal=3)
            numpy.testing.assert_almost_equal(radius_min, 0.054, decimal=3)
            print("Test passed: KIC catalog pull from Vizier using astroquery")

            (a, b), mass, mass_min, mass_max, radius, radius_min, radius_max = catalog_info(
                TIC_ID=279741377
            )
            numpy.testing.assert_equal((a, b), (0.354, 0.2321))
            numpy.testing.assert_equal(mass, 1.12)
            numpy.testing.assert_equal(mass_min, 0.148162)
            numpy.testing.assert_equal(mass_max, 0.148162)
            numpy.testing.assert_equal(radius, 3.08104)
            numpy.testing.assert_equal(radius_min, 0.167789)
            numpy.testing.assert_equal(radius_max, 0.167789)
            print("Test passed: TESS Input Catalog (TIC) pull from Vizier using astroquery")

            print("All tests passed")
        else:
            print('Skipping test, not working on Python 2')




from __future__ import division, print_function
import numpy
import batman
from foldedleastsquares import transitleastsquares


import unittest

class TestTransitDepthMin(unittest.TestCase):
    def test(self):
        print("Starting test: transit_depth_min...", end="")
        # Test for transit_depth_min=1000*10**-6, where no transit is found

        # Create test data
        start = 48
        days = 365.25 * 3
        samples_per_day = 12  # 48
        samples = int(days * samples_per_day)  # 48
        t = numpy.linspace(start, start + days, samples)

        # Use batman to create transits
        ma = batman.TransitParams()
        ma.t0 = (
            start + 20
        )  # time of inferior conjunction; first transit is X days after start
        ma.per = 365.25  # orbital period
        ma.rp = 6371 / 696342  # 6371 planet radius (in units of stellar radii)
        ma.a = 217  # semi-major axis (in units of stellar radii)
        ma.inc = 90  # orbital inclination (in degrees)
        ma.ecc = 0  # eccentricity
        ma.w = 90  # longitude of periastron (in degrees)
        ma.u = [0.5]  # limb darkening coefficients
        ma.limb_dark = "linear"  # limb darkening model
        m = batman.TransitModel(ma, t)  # initializes model
        original_flux = m.light_curve(ma)  # calculates light curve

        # Create noise and merge with flux
        ppm = 5
        stdev = 10 ** -6 * ppm
        noise = numpy.random.normal(0, stdev, int(samples))
        y = original_flux + noise
        y[1] = numpy.nan
        model = transitleastsquares(t, y)
        results = model.power(
            transit_depth_min=1000 * 10 ** -6,
            period_min=360,
            period_max=370,
            oversampling_factor=5,
            duration_grid_step=1.02,
            T0_fit_margin=0.1,
        )

        numpy.testing.assert_equal(results.transit_times, numpy.nan)
        numpy.testing.assert_equal(results.period, numpy.nan)
        numpy.testing.assert_equal(results.depth, 1)
        numpy.testing.assert_equal(results.duration, numpy.nan)
        numpy.testing.assert_equal(results.snr, numpy.nan)
        numpy.testing.assert_equal(results.snr_pink_per_transit, numpy.nan)
        numpy.testing.assert_equal(results.odd_even_mismatch, numpy.nan)
        numpy.testing.assert_equal(results.SDE, 0)
        numpy.testing.assert_equal(results.SDE_raw, 0)
        numpy.testing.assert_equal(results.in_transit_count, numpy.nan)
        numpy.testing.assert_equal(results.after_transit_count, numpy.nan)
        numpy.testing.assert_equal(results.before_transit_count, numpy.nan)
        numpy.testing.assert_almost_equal(results.chi2_min, 13148.0)
        numpy.testing.assert_almost_equal(results.chi2red_min, 1.0003043213633598)
        numpy.testing.assert_equal(len(results.periods), 278)
        numpy.testing.assert_almost_equal(max(results.periods), 369.9831654894093)
        numpy.testing.assert_almost_equal(min(results.periods), 360.0118189140635)
        numpy.testing.assert_almost_equal(max(results.power), 0)
        numpy.testing.assert_almost_equal(min(results.periods), 360.0118189140635)
        numpy.testing.assert_almost_equal(min(results.power), 0)
        numpy.testing.assert_almost_equal(max(results.chi2), 13148.0)
        numpy.testing.assert_almost_equal(max(results.chi2red), 1.0003043213633598)
        print("Test passed: transit_depth_min=1000*10**-6, where no transit is fit")



from __future__ import division, print_function
import numpy

from foldedleastsquares import DefaultTransitTemplateGenerator

import unittest

class TestPeriodGrid(unittest.TestCase):
    def test(self):
        print("Starting test: period_grid...", end="")
        default_transit_template_generator = DefaultTransitTemplateGenerator()
        periods = default_transit_template_generator.period_grid(R_star=1, M_star=1, time_span=0.1)
        numpy.testing.assert_almost_equal(max(periods), 2.4999999999999987)
        numpy.testing.assert_almost_equal(min(periods), 0.6002621413799498)
        numpy.testing.assert_equal(len(periods), 268)

        periods = default_transit_template_generator.period_grid(R_star=1, M_star=1, time_span=20)
        numpy.testing.assert_almost_equal(max(periods), 10)
        numpy.testing.assert_almost_equal(min(periods), 0.6015575922909607)
        numpy.testing.assert_equal(len(periods), 1716)

        periods = default_transit_template_generator.period_grid(
            R_star=5,  # R_sun
            M_star=1,  # M_sun
            time_span=20,  # days
            period_min=0,
            period_max=999,
            oversampling_factor=3,
        )
        numpy.testing.assert_almost_equal(max(periods), 10)
        numpy.testing.assert_almost_equal(min(periods), 0.6015575922909607)
        numpy.testing.assert_equal(len(periods), 1716)

        periods = default_transit_template_generator.period_grid(
            R_star=1,  # R_sun
            M_star=1,  # M_sun
            time_span=20,  # days
            period_min=0,
            period_max=999,
            oversampling_factor=3,
        )
        numpy.testing.assert_almost_equal(max(periods), 10)
        numpy.testing.assert_almost_equal(min(periods), 0.60155759)
        numpy.testing.assert_equal(len(periods), 1716)

        periods = default_transit_template_generator.period_grid(
            R_star=0.1,  # R_sun
            M_star=1,  # M_sun
            time_span=1000,  # days
            period_min=0,
            period_max=999,
            oversampling_factor=3,
        )
        numpy.testing.assert_equal(len(periods), 4308558)

        periods = default_transit_template_generator.period_grid(
            R_star=1,  # R_sun
            M_star=1,  # M_sun
            time_span=20,  # days
            period_min=0.5,
            period_max=999,
            oversampling_factor=3,
        )
        numpy.testing.assert_almost_equal(max(periods), 10)
        numpy.testing.assert_almost_equal(min(periods), 0.500, 3)
        numpy.testing.assert_equal(len(periods), 1895)
        print("passed")



from __future__ import division, print_function

import os
import time

import numpy
import scipy
import scipy.signal
from foldedleastsquares import transitleastsquares, transit_mask, cleaned_array
import pkg_resources

import unittest

class TestMultiPlanet(unittest.TestCase):
    def loadfile(self, filename):
        filename = TestMultiPlanet.get_path(filename)
        try:
            data = numpy.genfromtxt(filename, delimiter=",", dtype="f8, f8, f8", names=["t", "y", "dy"])
            return data["t"], data["y"], data["dy"]
        except:
            data = numpy.genfromtxt(filename, delimiter=",", dtype="f8, f8", names=["t", "y"])
            return data["t"], data["y"], None

    @staticmethod
    def get_path(path):
        """
        Gets right path for tests environment
        :param path:
        :return: the real path of the test resource
        """
        return pkg_resources.resource_filename(__name__, path)

    def test_gpu(self):
        print("Starting test: Multi-planet...", end="")
        t, y, dy = self.loadfile("EPIC201367065.csv")
        trend = scipy.signal.medfilt(y, 25)
        y_filt = y / trend
        model = transitleastsquares(t, y_filt)
        total_gpu_time = time.time()
        results_gpu = model.power(use_gpu=True, show_progress_bar=False)
        total_gpu_time = time.time() - total_gpu_time
        total_gpu_approximate_time = time.time()
        results_gpu_approximate = model.power(use_gpu=True, gpu_approximate=True, show_progress_bar=False)
        total_gpu_approximate_time = time.time() - total_gpu_approximate_time
        #self.plot_results(results_gpu, t ,y)
        total_cpu_time = time.time()
        results_cpu = model.power(use_threads=os.cpu_count(), show_progress_bar=False)
        total_cpu_time = time.time() - total_cpu_time
        #self.plot_results(results_cpu, t, y)
        print("TIMES CPU: ", total_cpu_time)
        print("TIMES GPU: ", total_gpu_time)
        print("TIMES GPU APPROXIMATE: ", total_gpu_approximate_time)
        #self.assertLess(total_gpu_time, total_cpu_time)
        self.assertLess(total_gpu_approximate_time, total_cpu_time)
        self.assertAlmostEqual(results_gpu.period, results_cpu.period, 3)
        self.assertAlmostEqual(results_gpu_approximate.period, results_cpu.period, 3)
        numpy.testing.assert_almost_equal(max(results_cpu.power), 47.585106066175584, decimal=3)
        numpy.testing.assert_almost_equal(
            max(results_cpu.power_raw), 42.93056655774114, decimal=3
        )
        numpy.testing.assert_almost_equal(min(results_cpu.power), -0.6531152137309356, decimal=3)
        numpy.testing.assert_almost_equal(
            min(results_cpu.power_raw), -0.3043720539933344, decimal=3
        )
        print("Detrending of power spectrum from power_raw passed")

        # Mask of the first planet
        intransit = transit_mask(t, results_cpu.period, 2 * results_cpu.duration, results_cpu.T0)
        y_second_run = y_filt[~intransit]
        t_second_run = t[~intransit]
        t_second_run, y_second_run = cleaned_array(t_second_run, y_second_run)

        # Search for second planet
        model = transitleastsquares(t_second_run, y_second_run)
        total_gpu_time = time.time()
        results_gpu = model.power(use_gpu=True, show_progress_bar=False)
        total_gpu_time = time.time() - total_gpu_time
        total_gpu_approximate_time = time.time()
        results_gpu_approximate = model.power(use_gpu=True, gpu_approximate=True, show_progress_bar=False)
        total_gpu_approximate_time = time.time() - total_gpu_approximate_time
        #self.plot_results(results_gpu, t_second_run, y_second_run)
        total_cpu_time = time.time()
        results_cpu = model.power(use_threads=os.cpu_count(), show_progress_bar=False)
        total_cpu_time = time.time() - total_cpu_time
        #self.plot_results(results_cpu, t_second_run, y_second_run)
        print("TIMES CPU: ", total_cpu_time)
        print("TIMES GPU: ", total_gpu_time)
        print("TIMES GPU APPROXIMATE: ", total_gpu_approximate_time)
        self.assertLess(total_gpu_time, total_cpu_time)
        self.assertLess(total_gpu_approximate_time, total_cpu_time)
        self.assertAlmostEqual(results_gpu.period, results_cpu.period, 3)
        self.assertAlmostEqual(results_gpu_approximate.period, results_cpu.period, 3)
        numpy.testing.assert_almost_equal(
            results_cpu.duration, 0.14596011421893682, decimal=3
        )
        numpy.testing.assert_almost_equal(
            results_cpu.SDE, 35.031824450265, decimal=3
        )
        numpy.testing.assert_almost_equal(
            results_cpu.rp_rs, 0.025852178872027086, decimal=3
        )

        print("Passed")

    def plot_results(self, results, t, y):
        import matplotlib.pyplot as plt
        import scipy.stats as stats
        plt.plot(results.periods, results.chi2)
        plt.show()
        plt.plot(results.periods, results.chi2red)
        plt.show()
        plt.plot(results.periods, results.power_raw)
        plt.show()
        plt.plot(results.periods, results.power)
        plt.show()
        time_folded = (t - results.T0) / results.period - numpy.floor((t - results.T0) / results.period)
        duration = results.duration / results.period
        args = numpy.argsort(time_folded).flatten()
        time_folded = time_folded[args]
        flux_folded = y[args]
        span = duration * 4
        bin_means, bin_edges, binnumber = stats.binned_statistic(time_folded, flux_folded, statistic='mean', bins=200)
        bin_stds, _, _ = stats.binned_statistic(time_folded, flux_folded, statistic='std', bins=200)
        bin_width = (bin_edges[1] - bin_edges[0])
        bin_centers = bin_edges[1:] - bin_width / 2
        plt.scatter(time_folded, flux_folded)
        plt.scatter(bin_centers, bin_means)
        plt.xlim(0.5 - span, 0.5 + span)
        plt.hlines(1, 0, 0.5 - duration / 2, color="red")
        plt.hlines(results.depth, 0.5 - duration / 2, 0.5 + duration / 2, color="red")
        plt.hlines(1, 0.5 + duration / 2, 1, color="red")
        plt.show()



from __future__ import division, print_function
import numpy
import batman
from foldedleastsquares import transitleastsquares


import unittest

class TestUncertainties(unittest.TestCase):
    def test(self):
        print("Starting test: uncertainties...", end="")
        numpy.random.seed(seed=0)  # reproducibility

        # Create test data
        start = 48
        days = 365.25 * 3
        samples_per_day = 12  # 48
        samples = int(days * samples_per_day)  # 48
        t = numpy.linspace(start, start + days, samples)

        # Use batman to create transits
        ma = batman.TransitParams()
        ma.t0 = (
            start + 20
        )  # time of inferior conjunction; first transit is X days after start
        ma.per = 365.25  # orbital period
        ma.rp = 6371 / 696342  # 6371 planet radius (in units of stellar radii)
        ma.a = 217  # semi-major axis (in units of stellar radii)
        ma.inc = 90  # orbital inclination (in degrees)
        ma.ecc = 0  # eccentricity
        ma.w = 90  # longitude of periastron (in degrees)
        ma.u = [0.5]  # limb darkening coefficients
        ma.limb_dark = "linear"  # limb darkening model
        m = batman.TransitModel(ma, t)  # initializes model
        original_flux = m.light_curve(ma)  # calculates light curve

        # Create noise and merge with flux
        ppm = 5
        stdev = 10 ** -6 * ppm
        noise = numpy.random.normal(0, stdev, int(samples))
        y = original_flux + noise

        # Inject excess noise near the end of the time series.
        # When searching without uncertainties, the SDE is 4.827
        # When searching with uncertainties, the SDE is larger, 5.254. Test passed!
        noise = numpy.random.normal(0, 10 * stdev, 3149)
        y[10000:] = y[10000:] + noise

        dy = numpy.full(len(y), stdev)
        dy[10000:] = 10 * stdev

        model = transitleastsquares(t, y, dy)
        results = model.power(
            period_min=360,
            period_max=370,
            oversampling_factor=3,
            duration_grid_step=1.05,
            T0_fit_margin=0.2,
        )
        numpy.testing.assert_almost_equal(results.SDE, 5.292594615900944, decimal=5)
        print("passed")



import math
import cupy
import numba
import numpy as np
import torch
from numba import cuda, float32

GPU_TPB = 32
DATA_LEN = 1000
PERIODS_LEN = 1
PATCHED_DATA_LEN = DATA_LEN + 1
DURATIONS_LEN = 30
MAX_DURATION = 5
TRANSIT_DEPTH_MIN = 1e-6
GPU_ORIGINAL_FLUX_LEN = 1000
T0_FIT_MARGIN = 0.01
G = 6.673e-11  # gravitational constant [m^3 / kg / s^2]
R_sun = 695508000  # radius of the Sun [m]
R_earth = 6371000  # radius of the Earth [m]
R_jup = 69911000  # radius of Jupiter [m]
M_sun = 1.989 * 10 ** 30  # mass of the Sun [kg]
SECONDS_PER_DAY = 86400
pi = 3.141592653589793
SIGNAL_DEPTH = 0.5


def print_device_info():
    gpu = cuda.get_current_device()
    print("name = %s" % gpu.name)
    print("maxThreadsPerBlock = %s" % str(gpu.MAX_THREADS_PER_BLOCK))
    print("maxBlockDimX = %s" % str(gpu.MAX_BLOCK_DIM_X))
    print("maxBlockDimY = %s" % str(gpu.MAX_BLOCK_DIM_Y))
    print("maxBlockDimZ = %s" % str(gpu.MAX_BLOCK_DIM_Z))
    print("maxGridDimX = %s" % str(gpu.MAX_GRID_DIM_X))
    print("maxGridDimY = %s" % str(gpu.MAX_GRID_DIM_Y))
    print("maxGridDimZ = %s" % str(gpu.MAX_GRID_DIM_Z))
    print("maxSharedMemoryPerBlock = %s" % str(gpu.MAX_SHARED_MEMORY_PER_BLOCK))
    print("asyncEngineCount = %s" % str(gpu.ASYNC_ENGINE_COUNT))
    print("canMapHostMemory = %s" % str(gpu.CAN_MAP_HOST_MEMORY))
    print("multiProcessorCount = %s" % str(gpu.MULTIPROCESSOR_COUNT))
    print("warpSize = %s" % str(gpu.WARP_SIZE))
    print("unifiedAddressing = %s" % str(gpu.UNIFIED_ADDRESSING))
    print("pciBusID = %s" % str(gpu.PCI_BUS_ID))
    print("pciDeviceID = %s" % str(gpu.PCI_DEVICE_ID))

def get_device_count():
    return cupy.cuda.runtime.getDeviceCount()

def change_data_len(value):
    global DATA_LEN
    DATA_LEN = value


def change_patched_data_len(value):
    global PATCHED_DATA_LEN
    PATCHED_DATA_LEN = value


def change_durations_len(value):
    global DURATIONS_LEN
    DURATIONS_LEN = value


def change_max_durations(value):
    global MAX_DURATION
    MAX_DURATION = value


def change_periods_len(value):
    global PERIODS_LEN
    PERIODS_LEN = value

def change_transit_min_depth(value):
    global TRANSIT_DEPTH_MIN
    TRANSIT_DEPTH_MIN = value

def change_t0_fit_margin(value):
    global T0_FIT_MARGIN
    T0_FIT_MARGIN = value


def get_data_len():
    return DATA_LEN

def get_patched_len():
    return PATCHED_DATA_LEN


def get_durations_len():
    return DURATIONS_LEN

def get_periods_len():
    return PERIODS_LEN

def gpu_compute_partitions(*arrays_args):
    used_bytes = [arrays_arg.size * 4 if arrays_args.dtype == np.float32 else arrays_args * 8
                  for arrays_arg in arrays_args]
    return gpu_compute_partitions_by_bytes(used_bytes)

def gpu_compute_partitions_by_bytes(periods_len, *byte_values):
    used_bytes = [arrays_arg for arrays_arg in byte_values]
    gp_mem_tuple = torch.cuda.mem_get_info()
    used_bytes = np.sum(used_bytes)
    partitions = used_bytes // gp_mem_tuple[0] + 1
    mem_per_partition = used_bytes / partitions
    return partitions, mem_per_partition

def compute_duration_partitions(durations_len, max_duration_in_samples):
    duration_tuples = [0]
    durations_len_boundary = 0
    block_size_threshold = 2**16 - 1
    while durations_len_boundary < durations_len:
        if durations_len_boundary + block_size_threshold // max_duration_in_samples < durations_len:
            durations_len_boundary = durations_len_boundary + block_size_threshold // max_duration_in_samples
        else:
            durations_len_boundary = durations_len
        duration_tuples = duration_tuples + [durations_len_boundary]
    return duration_tuples


def get_memory_mb():
    free_memory, all_memory = torch.cuda.mem_get_info()
    return free_memory // 1e6, all_memory // 1e6



device = True
inline = True


@cuda.jit(device=device, inline=inline)
def partition(arr, ids, l, h):
    """
    Partition using pivot.

    Function takes last element as pivot, places the pivot element at its correct
    position in sorted array, and places all smaller (smaller than pivot) to left of
    pivot and all greater elements to right of pivot

    Source: Modified from https://www.geeksforgeeks.org/iterative-quick-sort/
    which was contributed by Mohit Kumra.

    Parameters
    ----------
    arr : vector of floats
        The array to be sorted.
    ids : vector of ints
        The unsorted IDs corresponding to arr, in other words range(len(arr)).
    l : int
        Starting index for sorting.
    h : int
        Ending index for sorting.

    Returns
    -------
    int
        the new pivot?

    """
    # index of smaller element
    i = l - 1

    pivot = arr[h]

    for j in range(l, h):

        # If current element is smaller than or equal to pivot
        if arr[j] <= pivot:

            # increment index of smaller element
            i += 1
            arr[i], arr[j] = arr[j], arr[i]
            ids[i], ids[j] = ids[j], ids[i]

    arr[i + 1], arr[h] = arr[h], arr[i + 1]
    ids[i + 1], ids[h] = ids[h], ids[i + 1]

    return i + 1


@cuda.jit(device=device, inline=inline)
def quickSortIterative(arr, stack, ids):
    """
    Perform iterative quicksort on array and an unsorted ID list of the array.

    Source: Modified from https://www.geeksforgeeks.org/iterative-quick-sort/
    which was contributed by Mohit Kumra.

    Parameters
    ----------
    arr : vector of floats
        The array to be sorted.
    stack : vector of ints
        Array initialized with 0's
    ids : vector of ints
        The unsorted IDs corresponding to arr, in other words range(len(arr)).

    Returns
    -------
    None.

    """
    # low and high indices.
    l, h = (0, len(arr) - 1)
    # stack = [0] * size
    # ids = list(range(len(arr)))

    # initialize top of stack
    top = -1

    # fill ids with range(len(arr))
    for i in range(len(arr)):
        ids[i] = i
        stack[i] = 0

    # push initial values of l and h to stack
    top = top + 1
    stack[top] = l
    top = top + 1
    stack[top] = h

    # Keep popping from stack while is not empty
    while top >= 0:

        # Pop h and l
        h = stack[top]
        top = top - 1
        l = stack[top]
        top = top - 1

        # Set pivot element at its correct position in
        # sorted array
        p = partition(arr, ids, l, h)

        # If there are elements on left side of pivot,
        # then push left side to stack
        if p - 1 > l:
            top = top + 1
            stack[top] = l
            top = top + 1
            stack[top] = p - 1

        # If there are elements on right side of pivot,
        # then push right side to stack
        if p + 1 < h:
            top = top + 1
            stack[top] = p + 1
            top = top + 1
            stack[top] = h

@cuda.jit
def periods_durations_in_samples(periods, durations_in_samples, R_s_max, M_s_max, R_s_min, M_s_min, upper_limit,
                                 original_flux_len, time_span, periods_durations_in_samples):
    x, y = cuda.grid(2)
    if x < periods.shape[0] and y < durations_in_samples.shape[0]:
        period_days = periods[x]
        p = periods[x] * SECONDS_PER_DAY # seconds per day
        R_s_max = R_s_max * R_sun  # radius of the Sun [m]
        M_s_max = M_s_max * M_sun  # mass of the Sun [kg]
        R_s_min = R_s_min * R_sun  # radius of the Sun [m]
        M_s_min = M_s_min * M_sun  # mass of the Sun [kg]
        T14max = (R_s_max + 2 * R_jup) * ((4 * p) / (pi * G * M_s_max)) ** (1 / 3)
        duration_max = T14max / p
        if duration_max > upper_limit:
            duration_max = upper_limit
        T14min = R_s_min * ((4 * p) / (pi * G * M_s_min)) ** (1 / 3)
        duration_min = T14min / p
        if duration_min > upper_limit:
            duration_min = upper_limit
        no_of_transits_naive = time_span / period_days
        duration_max = int(math.ceil(duration_max * original_flux_len * ((no_of_transits_naive + 1) / no_of_transits_naive)))
        duration_min = int(math.floor(duration_min * original_flux_len))
        if durations_in_samples[y] >= duration_min and durations_in_samples[y] <= duration_max:
            periods_durations_in_samples[x, y] = int(durations_in_samples[y])

def periods_durations_in_samples_sim(periods, durations_in_samples, R_s_max, M_s_max, R_s_min, M_s_min, upper_limit,
                                 original_flux_len, time_span, periods_durations_in_samples, x, y):
    #x, y = cuda.grid(2)
    if x < periods.shape[0] and y < durations_in_samples.shape[0]:
        period_days = periods[x]
        p = period_days * SECONDS_PER_DAY # seconds per day
        R_s_max = R_s_max * R_sun  # radius of the Sun [m]
        M_s_max = M_s_max * M_sun  # mass of the Sun [kg]
        R_s_min = R_s_min * R_sun  # radius of the Sun [m]
        M_s_min = M_s_min * M_sun  # mass of the Sun [kg]
        T14min = R_s_min * ((4 * p) / (pi * G * M_s_min)) ** (1 / 3)
        T14max = (R_s_max + 2 * R_jup) * ((4 * p) / (pi * G * M_s_max)) ** (1 / 3)
        duration_max = T14max / p
        if duration_max > upper_limit:
            duration_max = upper_limit
        duration_min = T14min / p
        if duration_min > upper_limit:
            duration_min = upper_limit
        no_of_transits_naive = time_span / period_days
        duration_max = int(math.ceil(duration_max * original_flux_len * ((no_of_transits_naive + 1) / no_of_transits_naive)))
        duration_min = int(math.floor(duration_min * original_flux_len))
        if durations_in_samples[y] >= duration_min and durations_in_samples[y] <= duration_max:
            periods_durations_in_samples[x, y] = int(durations_in_samples[y])

@cuda.jit
def foldfast(periods, time, time_folded):
    x, y = cuda.grid(2)
    if x < time_folded.shape[0] and y < time_folded.shape[1]:
        time_folded[x][y] = time[y] / periods[x] - math.floor(time[y] / periods[x])

def foldfast_sim(periods, time, time_folded, x, y):
    if x < time_folded.shape[0] and y < time_folded.shape[1]:
        time_folded[x][y] = time[y] / periods[x] - math.floor(time[y] / periods[x])

@cuda.reduce
def gpu_sum_reduce(a, b):
    return a + b

@cuda.jit
def gpu_edge_effect_correction_old(folded_flux, folded_dy, folded_patched_flux_deviation_corrected,
                               folded_inverse_squared_patched_folded_dy,
                               folded_patched_flux, patched_folded_dy, edge_effect):
    x, y = cuda.grid(2)
    if x < folded_patched_flux.shape[0]:
        if y < folded_flux.shape[1]:
            folded_patched_flux[x][y] = folded_flux[x][y]
            patched_folded_dy[x][y] = folded_dy[x][y]
            folded_inverse_squared_patched_folded_dy[x][y] = 1 / folded_dy[x][y] ** 2
            folded_patched_flux_deviation_corrected[x][y] = (1 - folded_flux[x][y]) ** 2 * folded_dy[x][y]
        else:
            folded_patched_flux[x][y] = folded_flux[x][y - folded_flux.shape[1]]
            patched_folded_dy[x][y] = folded_dy[x][y - folded_flux.shape[1]]
            folded_inverse_squared_patched_folded_dy[x][y] = 1 / folded_dy[x][y - folded_flux.shape[1]] ** 2
            folded_patched_flux_deviation_corrected[x][y] = (1 - folded_flux[x][y - folded_flux.shape[1]]) ** 2 * \
                                                            folded_dy[x][y - folded_flux.shape[1]]
            cuda.atomic.add(edge_effect, x, ((1 - (folded_patched_flux[x][y])) ** 2) * folded_inverse_squared_patched_folded_dy[x][y])
        #folded_patched_zero_flux[x][y] = folded_patched_flux[x][y] - 1

@cuda.jit
def gpu_edge_effect_correction(folded_flux, folded_dy, oot_residuals,
                               folded_inverse_squared_patched_folded_dy,
                               folded_patched_flux, patched_folded_dy, edge_effect):
    x, y = cuda.grid(2)
    if x < folded_patched_flux.shape[0] and y < folded_patched_flux.shape[1]:
        if y < folded_flux.shape[1]:
            folded_patched_flux[x][y] = folded_flux[x][y]
            patched_folded_dy[x][y] = folded_dy[x][y]
            folded_inverse_squared_patched_folded_dy[x][y] = 1 / folded_dy[x][y] ** 2
            oot_residuals[x][y] = (1 - folded_flux[x][y]) ** 2 * \
                                  folded_inverse_squared_patched_folded_dy[x][y]
        else:
            #numpy.sum(((1 - patched_data) ** 2) * inverse_squared_patched_dy)
            folded_patched_flux[x][y] = folded_flux[x][y - folded_flux.shape[1]]
            patched_folded_dy[x][y] = folded_dy[x][y - folded_flux.shape[1]]
            folded_inverse_squared_patched_folded_dy[x][y] = 1 / patched_folded_dy[x][y] ** 2
            oot_residuals[x][y] = (1 - folded_patched_flux[x][y]) ** 2 * \
                                  folded_inverse_squared_patched_folded_dy[x][y]
            cuda.atomic.add(edge_effect, x, oot_residuals[x][y])
        #folded_patched_zero_flux[x][y] = folded_patched_flux[x][y] - 1

def gpu_edge_effect_correction_sim(folded_flux, folded_dy, folded_patched_flux_deviation_corrected,
                               folded_inverse_squared_patched_folded_dy,
                               folded_patched_flux, patched_folded_dy, edge_effect, x, y):
    #x, y = cuda.grid(2)
    if x < folded_patched_flux.shape[0]:
        if y < folded_flux.shape[1]:
            folded_patched_flux[x][y] = folded_flux[x][y]
            patched_folded_dy[x][y] = folded_dy[x][y]
            folded_inverse_squared_patched_folded_dy[x][y] = 1 / folded_dy[x][y] ** 2
            folded_patched_flux_deviation_corrected[x][y] = (1 - folded_flux[x][y]) ** 2 * \
                                                            folded_inverse_squared_patched_folded_dy[x][y]
        else:
            #numpy.sum(((1 - patched_data) ** 2) * inverse_squared_patched_dy)
            folded_patched_flux[x][y] = folded_flux[x][y - folded_flux.shape[1]]
            patched_folded_dy[x][y] = folded_dy[x][y - folded_flux.shape[1]]
            folded_inverse_squared_patched_folded_dy[x][y] = 1 / patched_folded_dy[x][y] ** 2
            folded_patched_flux_deviation_corrected[x][y] = (1 - folded_patched_flux[x][y]) ** 2 * \
                                                            folded_inverse_squared_patched_folded_dy[x][y]
            edge_effect[x] = ((1 - (folded_patched_flux[x][y])) ** 2) * folded_inverse_squared_patched_folded_dy[x][y]

@cuda.jit
def sum_value_to_array(array, value):
    x, y = cuda.grid(2)
    if x < array.shape[0] and y < array.shape[1]:
        array[x, y] = array[x, y] + value

@cuda.jit
def subtract_array_to_value(value, array):
    x, y = cuda.grid(2)
    if x < array.shape[0] and y < array.shape[1]:
        array[x, y] = value - array[x, y]

@cuda.jit
def multiply_by_value(array, value):
    x, y = cuda.grid(2)
    if x < array.shape[0] and y < array.shape[1]:
        array[x, y] = array[x, y] * value

@cuda.jit
def multiply_by_itself_position_keep_sign(array1, array3):
    x, y = cuda.grid(2)
    if x < array1.shape[0] and y < array1.shape[1]:
        if array1[x, y] < 0:
            array3[x, y] = - array1[x, y] * array1[x, y]
        else:
            array3[x, y] = array1[x, y] * array1[x, y]

@cuda.jit
def multiply_by_itself_position(array1, array3):
    x, y = cuda.grid(2)
    if x < array1.shape[0] and y < array1.shape[1]:
        array3[x, y] = array1[x, y] * array1[x, y]



@cuda.jit
def residuals_cauchy(patched_flux, patched_flux_pow, patched_flux_pow_signed, folded_patched_flux_deviation_corrected,
                     inverse_squared_patched_folded_dy,
                     periods_durations_in_samples, fullsum, folded_running_means, overshoots,
                     lc_arr_pow, lc_arr_pow_signed, residuals_data):
    x, z, y = cuda.grid(3)
    if x < patched_flux.shape[0] and y < periods_durations_in_samples.shape[1]:
        duration_in_samples = periods_durations_in_samples[x][y]
        if duration_in_samples > 0 and 0 < z < patched_flux.shape[1] - duration_in_samples - 1:
            cumsum_left_index = z - 1
            cumsum_right_index = cumsum_left_index + duration_in_samples
            mean = 1 - ((folded_running_means[x, cumsum_right_index] -
                    folded_running_means[x, cumsum_left_index]) / duration_in_samples)
            xth_point = int(duration_in_samples * T0_FIT_MARGIN)
            if xth_point < 1:
                xth_point = 1
            if mean > TRANSIT_DEPTH_MIN and z % xth_point == 0:
                target_depth = mean * overshoots[y]
                patched_flux_pow_sum = (
                            patched_flux_pow[x, cumsum_right_index] - patched_flux_pow[x, cumsum_left_index])
                patched_flux_pow_signed_sum = (patched_flux_pow_signed[x, cumsum_right_index] - patched_flux_pow_signed[
                    x, cumsum_left_index])
                avg_inverse_squared_patched_folded_dy = (inverse_squared_patched_folded_dy[x, cumsum_right_index]
                                                         - inverse_squared_patched_folded_dy[
                                                             x, cumsum_left_index]) / duration_in_samples
                # (patched_data - 1 + (1 - lc_arr) * rev_sc) ** 2 * inverse_sq_dy
                # (patched_data_norm ** 2 + ((1 - lc_arr) * rev_sc) ** 2 + 2 * patched_data_norm * (1 - lc_arr) * rev_sc) * inverse_sq_dy # assume avg invs sq dy
                # #patched_data_norm ** 2 out of the equation
                # ((1 - lc_arr) * rev_sc) ** 2 + 2 * patched_data_norm * (1 - lc_arr) * rev_sc
                # # (1 - lc_arr) * rev_sc becomes lc_arr_norm = (1 - lc_arr) / SIGNAL_DEPTH
                # (lc_arr_norm * target_depth) ** 2 + 2 * patched_data_norm * lc_arr_norm * target_depth
                # target_depth * ((lc_arr_norm_pw * target_depth) + 2 * patched_data_norm * lc_arr_norm)

                # target_depth * ((lc_pow ** 2 * target_depth) + 2 * patched_data_norm * lc_arr_norm)
                # cauchy inequality
                # 2 * target_depth_pow * sqrt(sum(patched_data_norm ** 2) * sum(lc_arr_norm** 2)  )
                flux_lc_arr_mult = patched_flux_pow_signed_sum * lc_arr_pow_signed[y, duration_in_samples - 1]
                flux_lc_arr_mult = math.sqrt(flux_lc_arr_mult) if flux_lc_arr_mult > 0 else - math.sqrt(
                    -flux_lc_arr_mult)
                current_stat = float32(
                    fullsum[x] -
                    (folded_patched_flux_deviation_corrected[x, cumsum_right_index]  # subtracting residuals from oot
                     - folded_patched_flux_deviation_corrected[x, cumsum_left_index])
                    # + avg_inverse_squared_patched_folded_dy * # adding intransit residuals
                    #     (patched_flux_pow_sum +
                    #      target_depth * (1 + 2 * (patched_flux_pow_sum * lc_arr_pow[y, duration_in_samples -1]) ** (1/2))
                    #     )
                    + avg_inverse_squared_patched_folded_dy * (
                            patched_flux_pow_sum +
                            target_depth * (
                                    lc_arr_pow[y, duration_in_samples - 1] * target_depth +
                                    2 * flux_lc_arr_mult
                            )
                    )
                )
                if current_stat < cuda.atomic.min(residuals_data, (x, 0), current_stat):
                    residuals_data[x, 0] = current_stat
                    residuals_data[x, 1] = y
                    residuals_data[x, 2] = 1 - target_depth

def gpu_residuals_cauchy_sim(patched_flux, patched_flux_pow, patched_flux_pow_signed, folded_patched_flux_deviation_corrected,
                  inverse_squared_patched_folded_dy,
                  periods_durations_in_samples, fullsum, folded_running_means, overshoots,
                  lc_arr, lc_arr_pow, lc_arr_pow_signed, residuals_data, x, y, z):
    #x, z, y = cuda.grid(3)
    if x < patched_flux.shape[0] and y < periods_durations_in_samples.shape[1]:
        duration_in_samples = periods_durations_in_samples[x][y]
        if duration_in_samples > 0 and 0 < z < patched_flux.shape[1] - duration_in_samples - 1:
            cumsum_left_index = z - 1
            cumsum_right_index = cumsum_left_index + duration_in_samples
            mean = float32(1 - ((folded_running_means[x, cumsum_right_index] -
                    folded_running_means[x, cumsum_left_index]) / float32(duration_in_samples)))
            xth_point = int(duration_in_samples * T0_FIT_MARGIN)
            if xth_point < 1:
                xth_point = 1
            if mean > TRANSIT_DEPTH_MIN and z % xth_point == 0:
                target_depth = mean * overshoots[y]
                patched_flux_pow_sum = (patched_flux_pow[x, cumsum_right_index] - patched_flux_pow[x, cumsum_left_index])
                patched_flux_pow_signed_sum = (patched_flux_pow_signed[x, cumsum_right_index] - patched_flux_pow_signed[x, cumsum_left_index])
                avg_inverse_squared_patched_folded_dy = (inverse_squared_patched_folded_dy[x, cumsum_right_index]
                                                         - inverse_squared_patched_folded_dy[x, cumsum_left_index]) / duration_in_samples
                # (patched_data - 1 + (1 - lc_arr) * rev_sc) ** 2 * inverse_sq_dy
                # (patched_data_norm ** 2 + ((1 - lc_arr) * rev_sc) ** 2 + 2 * patched_data_norm * (1 - lc_arr) * rev_sc) * inverse_sq_dy # assume avg invs sq dy
                # #patched_data_norm ** 2 out of the equation
                # ((1 - lc_arr) * rev_sc) ** 2 + 2 * patched_data_norm * (1 - lc_arr) * rev_sc
                # # (1 - lc_arr) * rev_sc becomes lc_arr_norm = (1 - lc_arr) / SIGNAL_DEPTH
                # (lc_arr_norm * target_depth) ** 2 + 2 * patched_data_norm * lc_arr_norm * target_depth
                # target_depth * ((lc_arr_norm_pw * target_depth) + 2 * patched_data_norm * lc_arr_norm)

                # target_depth * ((lc_pow ** 2 * target_depth) + 2 * patched_data_norm * lc_arr_norm)
                # cauchy inequality
                # 2 * target_depth_pow * sqrt(sum(patched_data_norm ** 2) * sum(lc_arr_norm** 2)  )
                flux_lc_arr_mult = patched_flux_pow_signed_sum * lc_arr_pow_signed[y, duration_in_samples - 1]
                flux_lc_arr_mult = math.sqrt(flux_lc_arr_mult) if flux_lc_arr_mult > 0 else - math.sqrt(
                    -flux_lc_arr_mult)
                current_stat = float32(
                    fullsum[x] -
                    (folded_patched_flux_deviation_corrected[x, cumsum_right_index] # subtracting residuals from oot
                     - folded_patched_flux_deviation_corrected[x, cumsum_left_index])
                    # + avg_inverse_squared_patched_folded_dy * # adding intransit residuals
                    #     (patched_flux_pow_sum +
                    #      target_depth * (1 + 2 * (patched_flux_pow_sum * lc_arr_pow[y, duration_in_samples -1]) ** (1/2))
                    #     )
                    + avg_inverse_squared_patched_folded_dy * (
                            patched_flux_pow_sum +
                            target_depth * (
                                    lc_arr_pow[y, duration_in_samples - 1] * target_depth +
                                    2 * flux_lc_arr_mult
                            )
                    )
                )
                if current_stat < residuals_data[x, 0]:
                    #residuals_data[x, 0] = current_stat
                    residuals_data[x, 1] = y
                    residuals_data[x, 2] = 1 - target_depth

@cuda.jit(fastmath=True)
def residuals(patched_flux, patched_flux_pow, folded_patched_flux_deviation_corrected,
              inverse_squared_patched_folded_dy,
              periods_durations_in_samples, fullsum, folded_running_means, overshoots,
              lc_arr, lc_arr_pow, residuals_data):
    x, z, y = cuda.grid(3)
    if x < patched_flux.shape[0] and y < periods_durations_in_samples.shape[1]:
        duration_in_samples = periods_durations_in_samples[x][y]
        if duration_in_samples > 0 and z > 0 and z < patched_flux.shape[1] - duration_in_samples:
            cumsum_left_index = z - 1
            cumsum_right_index = cumsum_left_index + duration_in_samples
            mean = 1 - ((folded_running_means[x, cumsum_right_index] -
                         folded_running_means[x, cumsum_left_index]) / float32(duration_in_samples))
            # xth_point = 1  # How many cadences the template shifts forward in each step
            # if duration_in_samples > T0_FIT_MARGIN:
            xth_point = int(duration_in_samples * T0_FIT_MARGIN)
            if xth_point < 1:
                xth_point = 1
            if mean > TRANSIT_DEPTH_MIN and z % xth_point == 0:
                target_depth = mean * overshoots[y]
                target_depth_pow = target_depth * target_depth
                duration_limit = z + duration_in_samples
                # oot residuals
                current_stat = float32(fullsum[x] - \
                                       (folded_patched_flux_deviation_corrected[
                                            x, cumsum_right_index]
                                        - folded_patched_flux_deviation_corrected[x, cumsum_left_index]))
                # average flux error within window
                avg_inverse_squared_patched_folded_dy = (inverse_squared_patched_folded_dy[x, duration_limit]
                                                         - inverse_squared_patched_folded_dy[x, z]) / \
                                                        duration_in_samples
                # window_stat = sum((pf + lc_arr * dep) ** 2) = sum(pf ** 2 + lcarr ** 2 * dep ** 2 + 2 * pf * lcarr * dep) =
                # sum(pf ** 2) + sum(lcarr ** 2 * dep ** 2) + sum(2 * pdf * lcarr * dep) =
                # sum(pf ** 2) + sum(lcarr ** 2 * dep ** 2) + 2 * dept * sum(pdf * lcarr)

                # window_stat = 0
                # for index_width in range(z, duration_limit):
                #     window_stat = window_stat + \
                #                   (patched_flux[x, index_width] + (lc_arr[y][index_width - z] * target_depth)) ** 2
                # current_stat = current_stat + window_stat * avg_inverse_squared_patched_folded_dy

                # window_stat = 0
                # for index_width in range(z, duration_limit):
                #     window_stat = window_stat + \
                #                   patched_flux[x, index_width] ** 2 + \
                #                   (lc_arr[y][index_width - z] * target_depth) ** 2 + \
                #                   2 * patched_flux[x, index_width] * lc_arr[y][index_width - z] * target_depth
                # current_stat = current_stat + window_stat * avg_inverse_squared_patched_folded_dy
                stat1 = float32(0)
                for index_width in range(z, duration_limit):
                    stat1 = stat1 + patched_flux[x, index_width] * lc_arr[y, index_width - z]
                current_stat = current_stat + avg_inverse_squared_patched_folded_dy * (
                        (patched_flux_pow[x, duration_limit - 1] - patched_flux_pow[x, z - 1]) +
                        lc_arr_pow[y, duration_in_samples - 1] * target_depth_pow +
                        2 * target_depth * stat1
                )
                if current_stat < cuda.atomic.min(residuals_data, (x, 0), current_stat):
                    residuals_data[x, 1] = y
                    residuals_data[x, 2] = 1 - target_depth

@cuda.jit(fastmath=True)
def gpu_residuals_per_period(x, patched_flux, patched_flux_pow, folded_patched_flux_deviation_corrected,
                  inverse_squared_patched_folded_dy,
                  periods_durations_in_samples, fullsum, folded_running_means, overshoots,
                  lc_arr, lc_arr_pow, residuals_data):
    z, y, d = cuda.grid(3)
    if x < patched_flux.shape[0] and y < periods_durations_in_samples.shape[1]:
        duration_in_samples = periods_durations_in_samples[x][y]
        if duration_in_samples > 0 and z > 0 and z < patched_flux.shape[1] - duration_in_samples:
            cumsum_left_index = z - 1
            cumsum_right_index = cumsum_left_index + duration_in_samples
            mean = 1 - ((folded_running_means[x, cumsum_right_index] -
                         folded_running_means[x, cumsum_left_index]) / float32(duration_in_samples))
            # xth_point = 1  # How many cadences the template shifts forward in each step
            # if duration_in_samples > T0_FIT_MARGIN:
            xth_point = int(duration_in_samples * T0_FIT_MARGIN)
            if xth_point < 1:
                xth_point = 1
            if mean > TRANSIT_DEPTH_MIN and z % xth_point == 0:
                target_depth = mean * overshoots[y]
                target_depth_pow = target_depth * target_depth
                duration_limit = z + duration_in_samples
                current_stat = float32(fullsum[x] - \
                                       (folded_patched_flux_deviation_corrected[
                                            x, cumsum_right_index]  # subtracting residuals from oot
                                        - folded_patched_flux_deviation_corrected[x, cumsum_left_index]))
                avg_inverse_squared_patched_folded_dy = (inverse_squared_patched_folded_dy[
                                                             x, duration_limit]  # subtracting residuals from oot
                                                         - inverse_squared_patched_folded_dy[
                                                             x, z]) / duration_in_samples
                # window_stat = sum((pf + lc_arr * dep) ** 2) = sum(pf ** 2 + lcarr ** 2 * dep ** 2 + 2 * pf * lcarr * dep) =
                # sum(pf ** 2) + sum(lcarr ** 2 * dep ** 2) + sum(2 * pdf * lcarr * dep) =
                # sum(pf ** 2) + sum(lcarr ** 2 * dep ** 2) + 2 * dept * sum(pdf * lcarr)

                # window_stat = 0
                # for index_width in range(z, duration_limit):
                #     window_stat = window_stat + \
                #                   (patched_flux[x, index_width] + (lc_arr[y][index_width - z] * target_depth)) ** 2
                # current_stat = current_stat + window_stat * avg_inverse_squared_patched_folded_dy

                # window_stat = 0
                # for index_width in range(z, duration_limit):
                #     window_stat = window_stat + \
                #                   patched_flux[x, index_width] ** 2 + \
                #                   (lc_arr[y][index_width - z] * target_depth) ** 2 + \
                #                   2 * patched_flux[x, index_width] * lc_arr[y][index_width - z] * target_depth
                # current_stat = current_stat + window_stat * avg_inverse_squared_patched_folded_dy
                stat1 = float32(0)
                for index_width in range(z, duration_limit):
                    stat1 = stat1 + patched_flux[x, index_width] * lc_arr[y, index_width - z]
                current_stat = current_stat + avg_inverse_squared_patched_folded_dy * (
                        (patched_flux_pow[x, duration_limit - 1] - patched_flux_pow[x, z - 1]) +
                        lc_arr_pow[y, duration_in_samples - 1] * target_depth_pow +
                        2 * target_depth * stat1
                )
                if current_stat < cuda.atomic.min(residuals_data, (x, 0), current_stat):
                    residuals_data[x, 0] = current_stat
                    residuals_data[x, 1] = y
                    residuals_data[x, 2] = 1 - target_depth

    def final_T0_fit(self, signal, depth, t, y, dy, period, T0_fit_margin, show_progress_bar):
        dur = len(signal)
        scale = tls_constants.SIGNAL_DEPTH / (1 - depth) if depth >= 0 else tls_constants.SIGNAL_DEPTH / (1 + depth)
        signal = [1 - ((1 - value) / scale) if value <= 1 else 1 + ((value - 1) / scale) for value in signal]
        samples_per_period = numpy.size(y)

        if T0_fit_margin == 0:
            points = samples_per_period
        else:
            step_factor = T0_fit_margin * dur
            points = int(samples_per_period / step_factor)
        if points > samples_per_period:
            points = samples_per_period

        # Create all possible T0s from the start of [t] to [t+period] in [samples] steps
        T0_array = numpy.linspace(
            start=numpy.min(t), stop=numpy.min(t) + period, num=points
        )

        # Avoid showing progress bar when expected runtime is short
        if points > tls_constants.PROGRESSBAR_THRESHOLD and show_progress_bar:
            show_progress_info = True
        else:
            show_progress_info = False

        residuals_lowest = float("inf")
        T0 = 0

        if show_progress_info:
            print("Searching for best T0 for period", format(period, ".5f"), "days")
            pbar2 = tqdm(total=numpy.size(T0_array))
        signal_ootr = numpy.ones(len(y[dur:]))

        # Future speed improvement possible: Add multiprocessing. Will be slower for
        # short data and T0_FIT_MARGIN > 0.01, but faster for large data with dense
        # sampling (T0_FIT_MARGIN=0)
        for Tx in T0_array:
            phases = fold(time=t, period=period, T0=Tx)
            sort_index = numpy.argsort(phases, kind="mergesort")  # 75% of CPU time
            phases = phases[sort_index]
            flux = y[sort_index]
            dy = dy[sort_index]

            # Roll so that the signal starts at index 0
            # Numpy roll is slow, so we replace it with less elegant concatenate
            # flux = numpy.roll(flux, roll_cadences)
            # dy = numpy.roll(dy, roll_cadences)
            roll_cadences = int(dur / 2) + 1
            flux = numpy.concatenate([flux[-roll_cadences:], flux[:-roll_cadences]])
            dy = numpy.concatenate([flux[-roll_cadences:], flux[:-roll_cadences]])

            residuals_intransit = numpy.sum((flux[:dur] - signal) ** 2 / dy[:dur] ** 2)
            residuals_ootr = numpy.sum((flux[dur:] - signal_ootr) ** 2 / dy[dur:] ** 2)
            residuals_total = residuals_intransit + residuals_ootr

            if show_progress_info:
                pbar2.update(1)
            if residuals_total < residuals_lowest:
                residuals_lowest = residuals_total
                T0 = Tx
        if show_progress_info:
            pbar2.close()
        return T0



from __future__ import division, print_function
import multiprocessing
import time

import numpy
import sys
import warnings
from functools import partial
try:
    import cupy
    import foldedleastsquares.gpu as gpu
    from foldedleastsquares.gpu import GPU_TPB
except Exception as e:
    print(e)
    print("Won't be able to use GPU")
from numba import cuda
from tqdm import tqdm
from .times import Times
from .template_generator.tailed_transit_template_generator import TailedTransitTemplateGenerator
from .template_generator.default_transit_template_generator import DefaultTransitTemplateGenerator
from . import tls_constants
from .core import (
    search_period,
)
from .validate import validate_inputs, validate_args



class transitleastsquares(object):
    """Compute the transit least squares of limb-darkened transit models"""

    def __init__(self, t, y, dy=None):
        self.t, self.y, self.dy = validate_inputs(t, y, dy)
        default_transit_template_generator = DefaultTransitTemplateGenerator()
        self.transit_template_generators = {"default": default_transit_template_generator,
                               "grazing": default_transit_template_generator,
                               "box": default_transit_template_generator,
                               "tailed": TailedTransitTemplateGenerator()}

    def power(self, **kwargs):
        """Compute the periodogram for a set of user-defined parameters"""

        print(tls_constants.TLS_VERSION)
        self, kwargs = validate_args(self, kwargs)
        transit_template_generator = self.transit_template_generators[self.transit_template]
        periods = self.period_grid if self.period_grid is not None else transit_template_generator.period_grid(
            R_star=self.R_star,
            M_star=self.M_star,
            time_span=numpy.max(self.t) - numpy.min(self.t),
            period_min=self.period_min,
            period_max=self.period_max,
            oversampling_factor=self.oversampling_factor,
            n_transits_min=self.n_transits_min,
        )

        durations = transit_template_generator.duration_grid(
            periods, shortest=1 / len(self.t), log_step=self.duration_grid_step
        )

        maxwidth_in_samples = int(numpy.max(durations) * numpy.size(self.y))
        if maxwidth_in_samples % 2 != 0:
            maxwidth_in_samples = maxwidth_in_samples + 1
        lc_cache_overview, lc_arr = transit_template_generator.get_cache(
            period_grid=periods,
            durations=durations,
            maxwidth_in_samples=maxwidth_in_samples,
            per=self.per,
            rp=self.rp,
            a=self.a,
            inc=self.inc,
            ecc=self.ecc,
            w=self.w,
            u=self.u,
            limb_dark=self.limb_dark,
        )

        print(
            "Searching "
            + str(len(self.y))
            + " data points, "
            + str(len(periods))
            + " periods from "
            + str(round(min(periods), 3))
            + " to "
            + str(round(max(periods), 3))
            + " days"
        )

        # Python 2 multiprocessing with "partial" doesn't work
        # For now, only single-threading in Python 2 is supported
        if sys.version_info[0] < 3:
            self.use_threads = 1
            warnings.warn("This TLS version supports no multithreading on Python 2")
        use_cpu = not self.use_gpu or 'gpu' not in sys.modules or gpu.get_device_count() == 0
        if use_cpu:
            if self.use_threads == multiprocessing.cpu_count():
                print("Using all " + str(self.use_threads) + " CPU threads")
            else:
                print(
                    "Using "
                    + str(self.use_threads)
                    + " of "
                    + str(multiprocessing.cpu_count())
                    + " CPU threads"
                )
        else:
            free_memory, all_memory = gpu.get_memory_mb()
            print(f"Using GPU with free {free_memory}/{all_memory} MB")
        if self.show_progress_bar:
            bar_format = "{desc}{percentage:3.0f}%|{bar}| {n_fmt}/{total_fmt} periods | {elapsed}<{remaining}"
            pbar = tqdm(total=numpy.size(periods), smoothing=0.3, bar_format=bar_format)

        if tls_constants.PERIODS_SEARCH_ORDER == "ascending":
            periods = numpy.flip(periods)
        elif tls_constants.PERIODS_SEARCH_ORDER == "descending":
            pass  # it already is
        elif tls_constants.PERIODS_SEARCH_ORDER == "shuffled":
            periods = numpy.random.permutation(periods)
        else:
            raise ValueError("Unknown PERIODS_SEARCH_ORDER")

        # Result lists now (faster), convert to numpy array later
        test_statistic_periods = []
        test_statistic_residuals = []
        test_statistic_rows = []
        test_statistic_depths = []
        st = time.time()
        if self.use_threads > 1 and use_cpu:  # Run multi-core search
            pool = multiprocessing.Pool(processes=self.use_threads)
            params = partial(
                search_period,
                transit_template_generator,
                t=self.t,
                y=self.y,
                dy=self.dy,
                transit_depth_min=self.transit_depth_min,
                R_star_min=self.R_star_min,
                R_star_max=self.R_star_max,
                M_star_min=self.M_star_min,
                M_star_max=self.M_star_max,
                lc_arr=lc_arr,
                lc_cache_overview=lc_cache_overview,
                T0_fit_margin=self.T0_fit_margin,
            )
            for data in pool.imap_unordered(params, periods):
                test_statistic_periods.append(data[0])
                test_statistic_residuals.append(data[1])
                test_statistic_rows.append(data[2])
                test_statistic_depths.append(data[3])
                if self.show_progress_bar:
                    pbar.update(1)
            pool.close()
        elif use_cpu:
            for period in periods:
                data = search_period(
                    transit_template_generator,
                    period=period,
                    t=self.t,
                    y=self.y,
                    dy=self.dy,
                    transit_depth_min=self.transit_depth_min,
                    R_star_min=self.R_star_min,
                    R_star_max=self.R_star_max,
                    M_star_min=self.M_star_min,
                    M_star_max=self.M_star_max,
                    lc_arr=lc_arr,
                    lc_cache_overview=lc_cache_overview,
                    T0_fit_margin=self.T0_fit_margin
                )
                test_statistic_periods.append(data[0])
                test_statistic_residuals.append(data[1])
                test_statistic_rows.append(data[2])
                test_statistic_depths.append(data[3])
                if self.show_progress_bar:
                    pbar.update(1)
        else:
            times = Times()
            times.begin_time("init")
            periods_f32 = numpy.array(periods).astype('float32')
            periods_len = len(periods_f32)
            gpu.change_periods_len(periods_f32)
            original_flux_len = len(self.y)
            time_span = max(self.t) - min(self.t)
            durations_in_samples = numpy.unique(lc_cache_overview["width_in_samples"])
            max_duration_in_samples = int(max(durations_in_samples))
            if max_duration_in_samples % 2 != 0:
                max_duration_in_samples = max_duration_in_samples + 1
            max_sample_signal_len_samples = max([len(signal) for signal in lc_arr])
            standardized_lc_arr = numpy.zeros((lc_arr.shape[0], max_sample_signal_len_samples), dtype=numpy.float32)
            for index, signal in enumerate(lc_arr):
                standardized_lc_arr[index][0:len(signal)] = (1 - lc_arr[index]) / tls_constants.SIGNAL_DEPTH
            standardized_lc_arr = cupy.asarray(standardized_lc_arr)
            standardized_lc_arr_pow = cupy.array(standardized_lc_arr)
            standardized_lc_arr_pow_signed = cupy.array(standardized_lc_arr)
            blockspergrid = ((periods_len + GPU_TPB) // GPU_TPB, (len(durations_in_samples) + GPU_TPB) // GPU_TPB)
            gpu.multiply_by_itself_position[blockspergrid, (GPU_TPB, GPU_TPB)] \
                (standardized_lc_arr, standardized_lc_arr_pow)
            gpu.multiply_by_itself_position_keep_sign[blockspergrid, (GPU_TPB, GPU_TPB)] \
                (standardized_lc_arr, standardized_lc_arr_pow_signed)
            cuda.synchronize()
            standardized_lc_arr_pow = cupy.cumsum(standardized_lc_arr_pow, axis=1)
            standardized_lc_arr_pow_signed = cupy.cumsum(standardized_lc_arr_pow_signed, axis=1)
            cuda.synchronize()
            periods_durations_in_samples = cuda.to_device(numpy.zeros((periods_len, len(durations_in_samples)), dtype=numpy.int32))
            gpu.periods_durations_in_samples[blockspergrid, (GPU_TPB, GPU_TPB)](cuda.to_device(periods_f32),
                                                                                cuda.to_device(durations_in_samples),
                                                                                self.R_star_max, self.M_star_max,
                                                                                self.R_star_min,
                                                                                self.M_star_min,
                                                                                tls_constants.FRACTIONAL_TRANSIT_DURATION_MAX,
                                                                                original_flux_len,
                                                                                time_span, periods_durations_in_samples)
            # dur_samp_host = periods_durations_in_samples.copy_to_host()
            # for x, period in enumerate(periods_f32):
            #     for y, duration in enumerate(dur_samp_host[x]):
            #         gpu.periods_durations_in_samples_sim(periods_f32, durations_in_samples,
            #                                                         self.R_star_max, self.M_star_max, self.R_star_min,
            #                                                         self.M_star_min,
            #                                                         tls_constants.FRACTIONAL_TRANSIT_DURATION_MAX,
            #                                                         original_flux_len,
            #                                                         time_span, dur_samp_host, x, y)
            cuda.synchronize()
            overshoots = cuda.to_device(lc_cache_overview["overshoot"].astype('float32'))
            gpu.change_data_len(original_flux_len)
            patched_data_len = gpu.get_data_len() + max_duration_in_samples
            gpu.change_patched_data_len(patched_data_len)
            gpu.change_transit_min_depth(self.transit_depth_min)
            time_f32 = cuda.to_device(self.t.astype('float32'))
            periods_f32 = cuda.to_device(periods_f32)
            partitions, mem_per_partition = gpu.gpu_compute_partitions_by_bytes(periods_len * 4, #periods_f32
                                                             original_flux_len * 4, #time_f32
                                                             periods_len * original_flux_len * 4 * 3, # time_folded
                                                             periods_len * original_flux_len * 4, # flux_folded
                                                             periods_len * original_flux_len * 4, # flux_folded take_along_aixs
                                                             periods_len * original_flux_len * 4, # dy_folded
                                                             periods_len * original_flux_len * 4, # dy_folded take_along_aixs
                                                             periods_len * original_flux_len * 8 * 3, # sort_indexes
                                                             periods_len * patched_data_len * 4, # folded_patched_flux
                                                             periods_len * patched_data_len * 4, # folded_patched_flux_pow
                                                             periods_len * patched_data_len * 4, # folded_patched_flux_pow cumsum
                                                             periods_len * patched_data_len * 4, # folded_patched_dy
                                                             periods_len * patched_data_len * 4, # folded_inverse_squared_patched_folded_dy
                                                             periods_len * patched_data_len * 4, # folded_inverse_squared_patched_folded_dy cumsum
                                                             periods_len * patched_data_len * 4, # folded_patched_flux_deviation_corrected
                                                             periods_len * patched_data_len * 4, # folded_patched_flux_deviation_corrected full_residuals_sum
                                                             periods_len * patched_data_len * 4, # folded_patched_flux_deviation_corrected cumsum
                                                             periods_len * 4, #edge_effect
                                                             periods_len * 4, #full_residuals_sum
                                                             periods_len * patched_data_len * 4, #folded_running_means
                                                             periods_len * len(durations_in_samples) * 4, #periods_durations_in_samples
                                                             periods_len * 3 * 4, #residuals_data
                                                             )
            partitions = int(partitions)
            cupy_memory_pool = cupy.get_default_memory_pool()
            with cupy.cuda.Device(0):
                cupy_memory_pool.set_limit(size=mem_per_partition)
            partition_periods_len = periods_len // partitions
            gpu.change_periods_len(partition_periods_len)
            gpu.change_durations_len(len(durations_in_samples))
            # Variables declaration
            residuals_data = numpy.full((periods_len, 3), original_flux_len, dtype=numpy.float32)
            partition_residuals_data_gpu = cuda.device_array((partition_periods_len, 3), dtype=numpy.float32)
            #zeros_patched_flux_prepend = cupy.asarray(numpy.zeros((partition_periods_len, 1), dtype=numpy.float32))
            partition_period_min = 0
            partitions = numpy.arange(0, partitions)
            times.end_time("init")
            for partition in partitions:
                times.begin_time(f"prep_{partition}")
                cupy_memory_pool.free_all_blocks()
                flux_folded = cupy.asarray(numpy.array([self.y] * partition_periods_len, dtype=numpy.float32))
                dy_folded = cupy.asarray(numpy.array([self.dy] * partition_periods_len, dtype=numpy.float32))
                time_folded = cupy.asarray(
                    numpy.zeros((partition_periods_len, gpu.get_data_len()), dtype=numpy.float32))
                cupy_memory_pool.free_all_blocks()
                # Partition logic
                partition_period_max = partition_period_min + partition_periods_len
                partition_period_max = partition_period_max if partition_period_max <= periods_len \
                    else periods_len
                cuda.synchronize()
                times.end_time(f"prep_{partition}")
                # Core logic
                times.begin_time(f"foldfast_{partition}")
                blockspergrid = ((gpu.get_periods_len() + GPU_TPB) // GPU_TPB,
                                 (gpu.get_data_len() + GPU_TPB) // GPU_TPB)
                gpu.foldfast[blockspergrid, (GPU_TPB, GPU_TPB)]\
                    (periods_f32[partition_period_min:partition_period_max], time_f32, time_folded)
                cuda.synchronize()
                cupy_memory_pool.free_all_blocks()
                times.end_time(f"foldfast_{partition}")
                times.begin_time(f"argsort_{partition}")
                sort_indexes = cupy.argsort(time_folded, axis=1)
                cupy_memory_pool.free_all_blocks()
                times.end_time(f"argsort_{partition}")
                times.begin_time(f"take_along_{partition}")
                flux_folded = cupy.take_along_axis(flux_folded, sort_indexes, axis=1)
                cupy_memory_pool.free_all_blocks()
                dy_folded = cupy.take_along_axis(dy_folded, sort_indexes, axis=1)
                cupy_memory_pool.free_all_blocks()
                cuda.synchronize()
                times.end_time(f"take_along_{partition}")
                times.begin_time(f"edge_effect_{partition}")
                folded_inverse_squared_patched_folded_dy = cupy.asarray(
                    numpy.zeros((partition_periods_len, gpu.get_patched_len()),
                                dtype=numpy.float32))
                folded_per_point_residuals = cupy.asarray(
                    numpy.zeros((partition_periods_len, gpu.get_patched_len()),
                                dtype=numpy.float32))
                folded_patched_flux = cupy.asarray(
                    numpy.zeros((partition_periods_len, gpu.get_patched_len()), dtype=numpy.float32))
                folded_patched_dy = cupy.asarray(
                    numpy.zeros((partition_periods_len, gpu.get_patched_len()), dtype=numpy.float32))
                edge_effect = cupy.asarray(numpy.zeros(partition_periods_len, dtype=numpy.float32))
                partition_residuals_data_gpu.copy_to_device(
                    numpy.full((partition_periods_len, 3), original_flux_len, dtype=numpy.float32))
                blockspergrid = ((gpu.get_periods_len() + GPU_TPB) // GPU_TPB,
                                 (gpu.get_patched_len() + GPU_TPB) // GPU_TPB)
                # flux = flux_folded.get()
                # dy = dy_folded.get()
                # dev = folded_patched_flux_deviation_corrected.get()
                # sq_dy = folded_inverse_squared_patched_folded_dy.get()
                # patched_flux = folded_patched_flux.get()
                # edge = edge_effect.get()
                # for x, period in enumerate(periods_f32[partition_period_min:partition_period_max]):
                #     for y in range(0, patched_data_len):
                #         gpu.gpu_edge_effect_correction_sim(flux, dy, dev,
                #              sq_dy,
                #              patched_flux, folded_patched_dy,
                #              edge, x, y)
                gpu.gpu_edge_effect_correction[blockspergrid, (GPU_TPB, GPU_TPB)]\
                    (flux_folded, dy_folded, folded_per_point_residuals,
                     folded_inverse_squared_patched_folded_dy,
                     folded_patched_flux, folded_patched_dy,
                     edge_effect)
                cuda.synchronize()
                del time_folded
                del flux_folded
                del dy_folded
                del sort_indexes
                cupy_memory_pool.free_all_blocks()
                times.end_time(f"edge_effect_{partition}")
                times.begin_time(f"full_residuals_{partition}")
                oot_sum = cupy.subtract(
                    cupy.sum(folded_per_point_residuals, axis=1),
                    edge_effect
                )
                del edge_effect
                cupy_memory_pool.free_all_blocks()
                times.end_time(f"full_residuals_{partition}")
                times.begin_time(f"means_{partition}")
                folded_running_means = cupy.cumsum(folded_patched_flux, axis=1)
                #folded_running_means = cupy.cumsum(cupy.concatenate((zeros_patched_flux_prepend, folded_patched_flux), axis=1), axis=1)
                cupy_memory_pool.free_all_blocks()
                times.end_time(f"means_{partition}")
                times.begin_time(f"sum_flux_{partition}")
                gpu.sum_value_to_array[blockspergrid, (GPU_TPB, GPU_TPB)](folded_patched_flux, -1)
                times.end_time(f"sum_flux_{partition}")
                times.begin_time(f"residuals_{partition}")
                folded_patched_flux_pow = cupy.asarray(
                    numpy.zeros((partition_periods_len, gpu.get_patched_len()), dtype=numpy.float32))
                folded_patched_flux_pow_signed = cupy.asarray(
                    numpy.zeros((partition_periods_len, gpu.get_patched_len()), dtype=numpy.float32))
                gpu.multiply_by_itself_position[blockspergrid, (GPU_TPB, GPU_TPB)]\
                    (folded_patched_flux, folded_patched_flux_pow)
                gpu.multiply_by_itself_position_keep_sign[blockspergrid, (GPU_TPB, GPU_TPB)]\
                    (folded_patched_flux, folded_patched_flux_pow_signed)
                cuda.synchronize()
                cupy_memory_pool.free_all_blocks()
                blockspergrid = ((gpu.get_periods_len() + GPU_TPB) // GPU_TPB,
                                 (gpu.get_patched_len() + GPU_TPB) // GPU_TPB,
                                 gpu.get_durations_len())
                if self.gpu_approximate:
                    gpu.residuals_cauchy[blockspergrid, (GPU_TPB, GPU_TPB, 1)]\
                        (folded_patched_flux,
                         cupy.cumsum(folded_patched_flux_pow, axis=1),
                         cupy.cumsum(folded_patched_flux_pow_signed, axis=1),
                         cupy.cumsum(folded_per_point_residuals, axis=1),
                         cupy.cumsum(folded_inverse_squared_patched_folded_dy, axis=1),
                         periods_durations_in_samples[partition_period_min:partition_period_max], oot_sum,
                         folded_running_means, overshoots,
                         standardized_lc_arr_pow,
                         standardized_lc_arr_pow_signed,
                         partition_residuals_data_gpu)
                else:
                    gpu.residuals[blockspergrid, (GPU_TPB, GPU_TPB, 1)] \
                        (folded_patched_flux,
                         cupy.cumsum(folded_patched_flux_pow, axis=1),
                         cupy.cumsum(folded_per_point_residuals, axis=1),
                         cupy.cumsum(folded_inverse_squared_patched_folded_dy, axis=1),
                         periods_durations_in_samples[partition_period_min:partition_period_max], oot_sum,
                         folded_running_means, overshoots,
                         standardized_lc_arr,
                         standardized_lc_arr_pow,
                         partition_residuals_data_gpu)
                # res_host = partition_residuals_data_gpu.copy_to_host()
                # dur_samp_host = periods_durations_in_samples[partition_period_min:partition_period_max].copy_to_host()
                # lc = standardized_lc_arr.get()
                # lc_pow = standardized_lc_arr_pow.get()
                # lc_pow_signed = standardized_lc_arr_pow_signed.get()
                # running = folded_running_means.get()
                # os = overshoots.copy_to_host()
                # sum = oot_sum.get()
                # inv_dy = cupy.cumsum(folded_inverse_squared_patched_folded_dy, axis=1).get()
                # corr = cupy.cumsum(folded_per_point_residuals, axis=1).get()
                # flux_pow = cupy.cumsum(folded_patched_flux_pow, axis=1).get()
                # flux_pow_signed = cupy.cumsum(folded_patched_flux_pow_signed, axis=1).get()
                # flux = folded_patched_flux.get()
                # for x, period in enumerate(periods_f32[partition_period_min:partition_period_max]):
                #     for y, duration in enumerate(dur_samp_host[x]):
                #         if duration > 0:
                #             for z in range(0, patched_data_len):
                #                 gpu.gpu_residuals_cauchy_sim \
                #                     (flux,
                #                      flux_pow, flux_pow_signed,
                #                      corr,
                #                      inv_dy,
                #                      dur_samp_host, sum,
                #                      running, os,
                #                      lc,
                #                      lc_pow, lc_pow_signed,
                #                      res_host, x, y, z)
                cuda.synchronize()
                del folded_patched_flux
                del folded_patched_flux_pow
                del folded_patched_flux_pow_signed
                del folded_per_point_residuals
                del folded_inverse_squared_patched_folded_dy
                del oot_sum
                del folded_running_means
                times.end_time(f"residuals_{partition}")
                times.begin_time(f"copy_{partition}")
                partition_residuals_data_gpu.copy_to_host(residuals_data[partition_period_min:partition_period_max])
                cuda.synchronize()
                cupy_memory_pool.free_all_blocks()
                times.end_time(f"copy_{partition}")
                if self.show_progress_bar:
                    pbar.update(partition_period_max - partition_period_min)
                partition_period_min = partition_period_max
            test_statistic_periods = periods_f32
            test_statistic_residuals = residuals_data[:, 0]
            test_statistic_rows = residuals_data[:, 1]
            test_statistic_depths = residuals_data[:, 2]
        total_time = time.time() - st
        #print("Total Time " + str(total_time / 60))
        test_statistic_periods = numpy.array(test_statistic_periods)
        sort_index = numpy.argsort(test_statistic_periods)
        test_statistic_periods = test_statistic_periods[sort_index]
        test_statistic_residuals = numpy.array(test_statistic_residuals)[sort_index]
        test_statistic_rows = numpy.array(test_statistic_rows)[sort_index]
        test_statistic_depths = numpy.array(test_statistic_depths)[sort_index]
        idx_best = numpy.argmin(test_statistic_residuals)
        best_row = int(test_statistic_rows[idx_best])
        duration = lc_cache_overview["duration"][best_row]
        maxwidth_in_samples = int(numpy.max(durations) * numpy.size(self.t))
        if max(test_statistic_residuals) == min(test_statistic_residuals) and len(test_statistic_residuals) > 1:
            no_transits_were_fit = True
            warnings.warn('No transit were fit. Try smaller "transit_depth_min"')
        else:
            no_transits_were_fit = False
        # Power spectra variants
        chi2 = test_statistic_residuals
        degrees_of_freedom = 4
        chi2red = test_statistic_residuals / (len(self.t) - degrees_of_freedom)
        chi2_min = numpy.min(chi2)
        chi2red_min = numpy.min(chi2red)
        return transit_template_generator.calculate_results(no_transits_were_fit, chi2, chi2red, chi2_min,
                                                            chi2red_min, test_statistic_periods, test_statistic_depths,
                                                            self, lc_arr, best_row, periods,
                                                            durations, duration, maxwidth_in_samples, len(self.y))




from __future__ import division, print_function
import numpy
from foldedleastsquares import transitleastsquares

import unittest

class TestValidation(unittest.TestCase):
    def test(self):

        t = y = numpy.linspace(0, 1, 2000)

        try:
            model = transitleastsquares(t, y)
            results = model.power(use_threads=0)
        except:
            print("Test passed: use_threads=0")
        try:
            model = transitleastsquares(t, y)
            results = model.power(use_threads="1")
        except:
            print("Test passed: use_threads='1'")



import time


class Times:
    times: dict = {}

    def begin_time(self, id):
        self.times[id] = time.time()

    def end_time(self, id):
        self.times[id] = time.time() - self.times[id]

    def get_times(self):
        return self.times

    def print_times(self, logger=None):
        if logger is not None:
            logger.info(self.times)
        else:
            print(self.times)



from __future__ import division, print_function

import numpy
from foldedleastsquares import FAP
import unittest

class TestFAP(unittest.TestCase):
    def test(self):
        print("Starting test FAP...", end="")
        numpy.testing.assert_equal(FAP(SDE=2), numpy.nan)
        numpy.testing.assert_equal(FAP(SDE=7), 0.009443778)
        numpy.testing.assert_equal(FAP(SDE=99), 8.0032e-05)
        print("passed")



from __future__ import division, print_function
import numpy
import scipy
import scipy.signal
from foldedleastsquares import transitleastsquares, transit_mask, cleaned_array
import pkg_resources

import unittest

class TestMultiPlanet(unittest.TestCase):
    def loadfile(self, filename):
        filename = TestMultiPlanet.get_path(filename)
        data = numpy.genfromtxt(filename, delimiter=",", dtype="f8, f8", names=["t", "y"])
        return data["t"], data["y"]

    @staticmethod
    def get_path(path):
        """
        Gets right path for tests environment
        :param path:
        :return: the real path of the test resource
        """
        return pkg_resources.resource_filename(__name__, path)

    def test(self):
        print("Starting test: Multi-planet...", end="")
        t, y = self.loadfile("EPIC201367065.csv")
        trend = scipy.signal.medfilt(y, 25)
        y_filt = y / trend

        model = transitleastsquares(t, y_filt)
        results = model.power()

        numpy.testing.assert_almost_equal(max(results.power), 44.990958555357125, decimal=3)
        numpy.testing.assert_almost_equal(
            max(results.power_raw), 42.93056655774114, decimal=3
        )
        numpy.testing.assert_almost_equal(min(results.power), -0.6175100139942419, decimal=3)
        numpy.testing.assert_almost_equal(
            min(results.power_raw), -0.3043720539933344, decimal=3
        )
        print("Detrending of power spectrum from power_raw passed")

        # Mask of the first planet
        intransit = transit_mask(t, results.period, 2 * results.duration, results.T0)
        y_second_run = y_filt[~intransit]
        t_second_run = t[~intransit]
        t_second_run, y_second_run = cleaned_array(t_second_run, y_second_run)

        # Search for second planet
        model_second_run = transitleastsquares(t_second_run, y_second_run)
        results_second_run = model_second_run.power()
        numpy.testing.assert_almost_equal(
            results_second_run.duration, 0.14596011421893682, decimal=3
        )
        numpy.testing.assert_almost_equal(
            results_second_run.SDE, 34.99113045986177, decimal=3
        )
        numpy.testing.assert_almost_equal(
            results_second_run.rp_rs, 0.025852178872027086, decimal=3
        )

        print("Passed")



import setuptools
from setuptools import setup
from os import path


exec(open(path.join("foldedleastsquares", 'version.py')).read())

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

    
setup(name='foldedleastsquares',
    version=TLS_VERSIONING,
    description='An optimized transit-fitting algorithm to search for periodic features in light curves',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/martindevora/tls',
    author='Martín Dévora Pajares',
    author_email='martin.devora.pajares@gmail.com',
    license='MIT',
    packages=setuptools.find_packages(),
    include_package_data=True,
    package_data={'': ['*.csv', '*.cfg']},
    install_requires=[
        'astropy==7.0.1',
        'astroquery==0.4.10',
        'numpy==2.2.5',
        'numba==0.61.2',
        'scipy==1.15.2',
        'tqdm',
        'batman-package==2.5.2',
        'argparse',
        'configparser',
        'torch==2.7.0'
        ],
    extras_require = {
        'cupy': '13.4.1'
    }
)



from __future__ import division, print_function

import unittest

import numpy
from foldedleastsquares.template_generator.default_transit_template_generator import DefaultTransitTemplateGenerator

class TestDurationGrid(unittest.TestCase):
    def test(self):
        print("Starting test: duration_grid...", end="")
        default_transit_template_generator = DefaultTransitTemplateGenerator()
        periods = default_transit_template_generator.period_grid(
            R_star=1,  # R_sun
            M_star=1,  # M_sun
            time_span=20,  # days
            period_min=0,
            period_max=999,
            oversampling_factor=3,
        )
        durations = DefaultTransitTemplateGenerator().duration_grid(periods, log_step=1.05, shortest=2)
        numpy.testing.assert_almost_equal(max(durations), 0.12)
        numpy.testing.assert_almost_equal(min(durations), 0.004562690993268325)
        numpy.testing.assert_equal(len(durations), 69)
        print("passed")
